"""A dict that transforms the keys before setting/retrieving values."""

import typing

KT = typing.TypeVar('KT')
TKT = typing.TypeVar('TKT', contravariant=True)
VT = typing.TypeVar('VT')


class TransformerDict(typing.Generic[KT, VT, TKT], typing.MutableMapping[TKT, VT]):  # noqa: WPS214
    def __init__(self, transformer: typing.Callable[[KT], TKT]) -> None:
        self._inner_dict: typing.Dict[TKT, VT] = {}
        self.transformer = transformer

    def __len__(self):
        return len(self._inner_dict)

    def __getitem__(self, key: KT) -> VT:
        return self._inner_dict[self.transformer(key)]

    def __setitem__(self, key: KT, value: VT) -> None:
        self._inner_dict[self.transformer(key)] = value

    def __delitem__(self, key: KT) -> None:  # noqa: WPS603
        del self._inner_dict[self.transformer(key)]

    def __iter__(self) -> typing.Iterator[TKT]:  # pragma: no cover
        return iter(self._inner_dict)

    # The following methods are overriden for type annotations
    def keys(self) -> typing.KeysView[TKT]:  # pragma: no cover
        return self._inner_dict.keys()

    def items(self):  # pragma: no cover
        return self._inner_dict.items()



"""Pre-condition check decorator."""

from asyncio import iscoroutinefunction
from typing import Any, Callable, Coroutine, Protocol, TypeVar, Union, cast

from asgiref.sync import async_to_sync
from makefun import wraps

Predicate = Union[Callable[..., bool], Callable[..., Coroutine[Any, Any, bool]]]

F = TypeVar('F', bound=Callable[..., Any])


class PreconditionDecorator(Protocol):  # pragma: no cover
    def __call__(self, fn: F) -> F:
        ...


def pre_condition(pre_condition_fn: Predicate) -> PreconditionDecorator:  # noqa: WPS231,WPS212
    """Executes a pre-condition function over the provided arguments before calling the decorated function.

    Automatically handles async pre-condition functions and wrapped functions.

    Exceptions raised in the pre-condition will pass through the decorator.
    If the pre-condition returns False, an UnmetPreconditionException will be raised.

    Args:
        pre_condition_fn (Predicate): The pre-condition function.

    Returns:
        PreconditionDecorator: The wrapped function
    """

    def pre_condition_decorator(fn: F) -> F:  # noqa: WPS231, WPS212
        if iscoroutinefunction(fn) and iscoroutinefunction(pre_condition_fn):

            # iscoroutinefunction doesn't work as a type guard
            async_precondition = cast(Callable[..., Coroutine[Any, Any, bool]], pre_condition_fn)

            @wraps(fn)  # type: ignore - The type markings are incorrect
            async def async_fn_with_async_precondition(*args: Any, **kwargs: Any):
                if not await async_precondition(*args, **kwargs):
                    raise UnmetPreconditionException
                return await fn(*args, **kwargs)

            return cast(F, async_fn_with_async_precondition)

        elif iscoroutinefunction(fn):

            @wraps(fn)  # type: ignore - The type markings are incorrect
            async def async_fn_with_sync_precondition(*args: Any, **kwargs: Any):  # noqa: WPS440
                if not pre_condition_fn(*args, **kwargs):
                    raise UnmetPreconditionException
                return await fn(*args, **kwargs)

            return cast(F, async_fn_with_sync_precondition)

        elif iscoroutinefunction(pre_condition_fn):

            @wraps(fn)  # type: ignore - The type markings are incorrect
            def sync_fn_with_async_precondition(*args: Any, **kwargs: Any):  # noqa: WPS440
                # Use async_to_sync to call an async from a sync: https://www.aeracode.org/2018/02/19/python-async-simplified/
                if not async_to_sync(pre_condition_fn)(*args, **kwargs):
                    raise UnmetPreconditionException
                return fn(*args, **kwargs)

            return cast(F, sync_fn_with_async_precondition)

        @wraps(fn)  # type: ignore - The type markings are incorrect
        def sync_fn_with_sync_precondition(*args: Any, **kwargs: Any):  # noqa: WPS440
            if not pre_condition_fn(*args, **kwargs):
                raise UnmetPreconditionException
            return fn(*args, **kwargs)

        return cast(F, sync_fn_with_sync_precondition)

    return pre_condition_decorator


class UnmetPreconditionException(Exception):
    ...



"""Feature flag management."""

import re
import warnings
from enum import Enum
from typing import Dict, Literal, Optional, Union

from outcome.utils import env
from outcome.utils.config import Config
from rich.console import Console
from rich.table import Table

StateSource = Literal['config', 'default']


class FeatureException(Exception):
    ...


class FeatureType(Enum):
    boolean = 'boolean'
    string = 'string'


_feature_pattern = r'^[a-z]+(_?[a-z]+_?[a-z])*$'
_feature_prefix = 'WITH_FEAT_'

_features: Dict[str, Union[bool, str, None]]
_feature_types: Dict[str, FeatureType]
_config: Config


def reset():
    global _config, _features, _feature_types
    _config = Config()  # noqa: WPS122,WPS442
    _features = {}  # noqa: WPS122,WPS442
    _feature_types = {}  # noqa: WPS122,WPS442


reset()


def _is_valid_feature_name(feature: str):
    return all(re.match(_feature_pattern, part) is not None for part in feature.split('.'))


def _feature_to_config_key(feature: str) -> str:
    replaced = feature.replace('.', '_').upper()
    return f'{_feature_prefix}{replaced}'


def _coerce_boolean(v: object) -> bool:
    return str(v).lower() in {'yes', 'y', 'true', 't', '1'}


def set_config(config: Config):
    global _config
    _config = config  # noqa: WPS122,WPS442


def register_feature(
    feature: str, default: Optional[Union[bool, str]] = None, feature_type: FeatureType = FeatureType.boolean,
) -> None:
    if not _is_valid_feature_name(feature):
        raise FeatureException(f'Invalid feature name: {feature}')

    if feature in _features:
        raise FeatureException(f'Duplicate feature: {feature}')

    if feature_type not in list(FeatureType):
        raise FeatureException(f'Invalid Type: {feature_type}')

    if default is None and feature_type == FeatureType.boolean:
        default = False

    _features[feature] = default
    _feature_types[feature] = feature_type


def features() -> Dict[str, bool]:
    return {k: is_active(k) for k in _features.keys()}


def display_features():  # pragma: no cover
    console = Console()

    table = Table(show_header=True, header_style='bold')
    table.add_column('Feature Flag')
    table.add_column('State', justify='right')
    table.add_column('Feature Type')
    table.add_column('Default State', justify='right')
    table.add_column('Config Key', justify='left')
    table.add_column('Set By', justify='right')

    def state_repr(state: Union[str, bool, None]):
        return '[bold green]active[/bold green]' if state else '[bold red]inactive[/bold red]'

    def type_repr(feature_type: FeatureType):
        return feature_type.value

    for feat, state in features().items():
        table.add_row(
            feat,
            state_repr(state),
            type_repr(_feature_types[feat]),
            state_repr(_features[feat]),
            _feature_to_config_key(feat),
            _feature_state_source(feat),
        )

    console.print(table)


def is_active(feature: str) -> bool:
    if not _feature_check(feature):
        return False

    value = _feature_status_from_config(feature)

    if value is not None:
        return value

    return bool(_features[feature])


def value(feature: str) -> Optional[Union[str, bool]]:
    if not _feature_check(feature):
        return None

    return _features[feature]


def _feature_status_from_config(feature: str) -> Optional[bool]:
    try:
        feature_key = _feature_to_config_key(feature)
        val = _config.get(feature_key)

        if _feature_types[feature] == FeatureType.boolean:
            return val if isinstance(val, bool) else _coerce_boolean(val)
        return bool(val)

    except KeyError:
        return None


def _feature_state_source(feature: str) -> StateSource:
    _feature_check(feature)

    val = _feature_status_from_config(feature)

    if val is None:
        return 'default'
    return 'config'


def _feature_check(feature: str):
    if feature not in _features:
        if env.is_dev():
            warnings.warn(
                f'Checking unknown feature "{feature}", maybe you forgot to register it? This will raise an exception in production',  # noqa: E501
                RuntimeWarning,
            )
            return False
        raise FeatureException(f'Unknown feature: {feature}')
    return True


def set_feature_default(feature: str, default_state: bool) -> None:
    _feature_check(feature)
    _features[feature] = default_state



"""Helpers to manipulate provider states headers."""

from __future__ import annotations

from typing import Dict, List, Optional

from pydantic import BaseModel


class ProviderState(BaseModel):
    method: str
    path: str
    state: str

    @classmethod
    def create(cls, method: str, path: str, state: str) -> ProviderState:
        return ProviderState(method=method, path=path, state=state)


class ProviderStates(BaseModel):
    __root__: List[ProviderState]


class ProviderStateMap:
    states: List[ProviderState]

    def __init__(self, states: Optional[List[ProviderState]] = None):
        self.states = []
        if states:
            self.add_states(states)

    def add_state(self, state: ProviderState) -> None:
        self.states.append(state)

    def add_states(self, states: List[ProviderState]) -> None:
        for state in states:
            self.add_state(state)

    def as_headers(self, key: str = 'pact-provider-state-map') -> Dict[str, str]:
        return {key: ProviderStates(__root__=self.states).json()}
